/*
 * Copyright (c) 2005 DocJava, Inc. All Rights Reserved.
 */
package math.fourierTransforms.pfa;

import j2d.ImageUtils;
import j2d.ShortImageBean;
import math.Mat2;
import math.complex.ComplexFloat1d;

import java.awt.*;
import java.awt.image.MemoryImageSource;

public class FFTImage {

    private int intImage[];
    private int imageWidth, imageHeight;
    // image.length, or imageWidth * imageHeight
    private int N;
    // scale is used to scale the FFT input to prevent overflow,
    // N = imageWidth * imageHeight is often used.
    private int scale;

    private boolean fftShift;

    private short alpha[];

    ComplexFloat1d red = new ComplexFloat1d();
    ComplexFloat1d blue = new ComplexFloat1d();
    ComplexFloat1d green = new ComplexFloat1d();
    //private final MyImageObserverUtils imageObserverUtils = new MyImageObserverUtils();


    public FFTImage(Image img) {
        initImage(img);
    }

    /**
     * uses one image to filter another.
     * only works if the two images are the same size, and if
     * the kernal is a mask defined in frequency space.
     */
    public static Image filter(Image img1, Image kernal) {
        FFTImage fimg1 = new FFTImage(img1);
        fimg1.fft();
        fimg1.scaleAndMask(kernal);
        fimg1.ifft();
        return fimg1.getImage();
    }

    public Image getImage() {
        Toolkit tk = Toolkit.getDefaultToolkit();
        return tk.createImage(
                new MemoryImageSource(
                        imageWidth,
                        imageHeight,
                        tk.getColorModel(),
                        intImage, 0,
                        imageWidth));
    }

    private void initImage(Image img) {
        ShortImageBean sib = new ShortImageBean(img);
        imageWidth = sib.getWidth();
        imageHeight = sib.getHeight();
        initVariables(sib.getPels(), imageWidth, true);
    }

    //todo test the below...it is untested!
    /**
     * The kernal is a filter mask!
     * Just screen out the pixels in frequency space. The input
     * kernal pixels are 0 or 255. So normalize to range from 0 to 1.
     * Do this by dividing by 255!
     *
     */
    public void scaleAndMask(Image kernal) {
        ShortImageBean sib = new ShortImageBean(kernal);
        red.scaleAndMask(sib.getR());
        green.scaleAndMask(sib.getG());
        blue.scaleAndMask(sib.getB());
    }


    public FFTImage(int intImage[], int imageWidth,
                    boolean fftShift) {
        initVariables(intImage, imageWidth, fftShift);

    }

    private void initVariables(int[] intImage, int imageWidth, boolean fftShift) {
        this.intImage = intImage;
        this.imageWidth = imageWidth;
        N = intImage.length;
        imageHeight = N / imageWidth;
        this.fftShift = fftShift;
        scale = N;

        alpha = ImageUtils.getAlpha(intImage);
        short r[] = ImageUtils.getRed(intImage);
        final int n = r.length;
        // If fftShift is true, shift the zero frequency to the center.
        red.setRe(fftReorder(r));
        green.setRe(fftReorder(ImageUtils.getGreen(intImage)));
        blue.setRe(fftReorder(ImageUtils.getBlue(intImage)));

        red.setIm(new float[n]);
        green.setIm(new float[n]);
        blue.setIm(new float[n]);
    }

    public void fft() {
        intImage = getFftIntArray();
    }

    public int[] getFftIntArray() {
        new FFT2d(red, imageWidth);
        new FFT2d(green, imageWidth);
        new FFT2d(blue, imageWidth);

        float magScale = 100;
        float resultRed[] = red.logScaleMagnitude(magScale);
        float resultGreen[] = green.logScaleMagnitude(magScale);
        float resultBlue[] = blue.logScaleMagnitude(magScale);

        return ImageUtils.getArgbToInt(alpha, resultRed,
                resultGreen, resultBlue);
    }

    public void ifft() {
        intImage = getIfftIntArray();
    }

    public int[] getIfftIntArray() {
        new IFFT2d(red, imageWidth);
        new IFFT2d(green, imageWidth);
        new IFFT2d(blue, imageWidth);

        return ImageUtils.getArgbToInt(alpha,
                ifftReorder(red.getRe()),
                ifftReorder(green.getRe()),
                ifftReorder(blue.getRe()));
    }

    // reorder color data of transforms.fft input.
    // 1. Convert color data from short to float.
    // 2. Scale the color data by scale.
    // 3. If fftShift is true, shift the zero frequency in the center of matrix.
    private float[] fftReorder(short color[]) {
        float result[] = new float[N];

        if (!fftShift) {   // Without zero frequency shift.
            for (int i = 0; i < N; i++)
                result[i] = (float) color[i] / scale;
        } else {            // With zero frequency shift.
            int k = 0;
            float alternateSign;
            for (int i = 0; i < imageHeight; i++)
                for (int j = 0; j < imageWidth; j++) {
                    alternateSign = ((i + j) % 2 == 0) ? -1 : 1;
                    result[i * imageWidth + j] = (color[k++] * alternateSign / scale);
                }
        }
        return result;
    } // End of function fftReorder().

    private short[] ifftReorder(float re[]) {
        short result[] = new short[N];

        if (!fftShift) {   // Without zero frequency shift.
            for (int i = 0; i < N; i++)
                result[i] = (short) (re[i] * scale);
        } else {            // With zero frequency shift.
            int k = 0;
            float alternateSign;
            for (int i = 0; i < imageHeight; i++)
                for (int j = 0; j < imageWidth; j++) {
                    alternateSign = ((i + j) % 2 == 0) ? -1 : 1;
                    result[i * imageWidth + j] = (short) (re[k++] * alternateSign * scale);
                }
        }
        return result;
    } // End of function fftReorder().


    public static void main(String[] args) {
        final int n = 4;
        ShortImageBean sib = new ShortImageBean(n, n);
        sib.setR(Mat2.getIdentity(n, n));
        sib.setG(Mat2.getIdentity(n, n));
        sib.setB(Mat2.getIdentity(n, n));
        sib.print();
         final int width = sib.getWidth();
        FFTImage fftimage = new FFTImage(sib.getPels(),
                width, true);

        final int height = sib.getHeight();
        Image psdImage = ImageUtils.getImage(
                fftimage.getFftIntArray(), width, height);
        ShortImageBean sibPsd = new ShortImageBean(psdImage);
        System.out.println("psd of identity");
        sibPsd.print();
        Image filteredImage = ImageUtils.getImage(
                fftimage.getIfftIntArray(), width, height);
        ShortImageBean sibF = new ShortImageBean(filteredImage);
        sibF.print();
    }

    private static void imageTest() {
        Image img = ImageUtils.getImage();
        imageTest(img);
    }

    private static void imageTest(Image img) {
        ImageUtils.displayImage(img, "original Image");
        ShortImageBean sib = new ShortImageBean(img);
        final int width = sib.getWidth();
        FFTImage fftimage = new FFTImage(sib.getPels(),
                width, true);
        final int height = sib.getHeight();
        Image psdImage = ImageUtils.getImage(
                fftimage.getFftIntArray(), width, height);
        ImageUtils.displayImage(psdImage, "psd image");
        Image filteredImage = ImageUtils.getImage(
                fftimage.getIfftIntArray(), width, height);
        ImageUtils.displayImage(filteredImage, "filtered image");
    }

} // End of class j2d.filters.fftConvolution.FFTImageAM.
